unit PayCashShop;

//***************************************************************************//
// (C) 2000 PayCash.                                          www.paycash.ru //
//           //
// WalletConnect.dll                                                         //
//***************************************************************************//

interface

uses Windows,Messages,Classes,SysUtils,PayCashTypes;

type
  TShopError = class ( Exception );
  TErrorEvent = procedure ( Sender : TObject; Mess : string ) of object;

  TPayCashShop = class;
  TContract = class;

  TMoneyRequest = record
    ErrorCode: TProcResult;
    Cypher: TCypher;
    ContractModifier: TParamsRecord;
    PayerAddress: PChar;
    RequestMode: word;
    ContractData: TContractData;
    ReplyResource: TReplyResource;
    end;

  TIncomingPaymentApproval=record
    ErrorCode: TProcResult;
    PaymentInfo: TIncomingPaymentInfo;
    ShortDescription:TShortDescription; // in/out
    ConfirmAction: TConfirmAction;
    BankConnection: DWord;
    ReplyResource: TReplyResource;
    end;

  TPaymentAuthorizationResult=record
    PaymentAuthorizationInfo:TPaymentAuthorizationInfo;
    ReplyResource: TReplyResource;
    end;

  TGetMoneyRequestProc = procedure (
    const ErrorCode: word;
    const ContractModifier: TStringList;
    out   RequestMode: word;
          Contract : TContract;
    out   Reply: string)of object;

  TGetIncomingPaymentApprovalProc=procedure (
    const ErrorCode: word;
    const ID : string;
    out   ConfirmAction : TConfirmAction ;
    out   Reply: string)of object;

  TPaymentAuthorizationResultProc=procedure(
    const ErrorCode : word;
    const ID : string;
    out   Reply: string ) of object;

  TContract = class
  private
    FPriceCount : integer;
    FAccount: int64;
    FBank: integer;
    FText: string;
    FDesc: string;
    FData: string;
    FPrices : packed array [0..MaxContractCurrencyOptionsCount - 1]
      of TCurrencyAndSum;
  public
    property Description : string read FDesc write FDesc;
    property Text : string read FText write FText;
    property Bank : integer read FBank write FBank;
    property Account : int64 read FAccount write FAccount;
    property Data : string read FData write FData;
    procedure AddPrice ( Cur : TCurrencyCode ; Price : TPreciseSum );
    end;

//    WalletConnect.DLL   
  TShopStartListen = function (Name:PChar; const ShopProcs:TShopProcsTable):
    DWord; stdcall;
  TShopStopListen = procedure ; stdcall;
  TShopGetMem = function (Size:DWord):pointer;
  TShopFreeMem = procedure(Mem:pointer; Size:DWord);
  TDateTimeToNanoTime=function(DateTime:TDateTime; var NanoTime: TNanoTime):
    DWord; stdcall;
  TNanoTimeToDateTime = function(NanoTime:TNanoTime; var DateTime:TDateTime):
    DWord; stdcall;
  TSystemTimeToNanoTime = function(SystemTime:TSystemTime;
    var NanoTime:TNanoTime):DWord; stdcall;
  TNanoTimeToSystemTime = function (NanoTime:TNanoTime;
    var SystemTime:TSystemTime):DWord; stdcall;

//     DLL 
  TWalletConnect = record
    Handle : cardinal;
    ShopStartListen : TShopStartListen;
    ShopStopListen : TShopStopListen;
    ShopGetMem : TShopGetMem;
    ShopFreeMem : TShopFreeMem;
    DateTimeToNanoTime : TDateTimeToNanoTime;
    NanoTimeToDateTime : TNanoTimeToDateTime;
    SystemTimeToNanoTime : TSystemTimeToNanoTime;
    NanoTimeToSystemTime : TNanoTimeToSystemTime;
    end;

//  TPayCashShop

  TPayCashShop = class ( TComponent )
  private
    FActive: boolean;
    FName: string;
    FWallet : TWalletConnect;
    FOnGetMoneyRequest: TGetMoneyRequestProc;
    FOnGetIncomingPaymentApproval: TGetIncomingPaymentApprovalProc;
    FOnPaymentAuthorizationResult: TPaymentAuthorizationResultProc;
    FOnActivate: TNotifyEvent;
    FOnDeactivate: TNotifyEvent;
    FOnError: TErrorEvent;
    FBankID: Cardinal;
    FAccountId: Int64;
    FDelay: integer;
    FSection : _RTL_CRITICAL_SECTION;
    procedure SetActive(const Value: boolean);
    procedure SetShopName(const Value: string);
    procedure GetMoneyRequest(var M : TMoneyRequest);
    procedure GetIncomingPaymentApproval(var M : TIncomingPaymentApproval);
    procedure PaymentAuthorizationResult(var M : TPaymentAuthorizationResult);
  protected
    procedure DoActivate;
    procedure DoDeactivate;
    procedure DoError( Mess : string );
  public
    procedure Connect;virtual;
    procedure Disconnect;virtual;
    constructor Create ( AOwner : TComponent );override;
    destructor Destroy;override;
    function  GetMem  ( Size : integer ): Pointer;
    procedure FreeMem ( Mem : Pointer; Size : integer );
    function  ShopString ( src : string ) : Pointer;
    procedure MakeReplyResource ( Source : string ;
          var Resource : TReplyResource);
  published
    property Active : boolean read FActive write SetActive;
    property ShopName : string read FName write SetShopName;
    property OnGetMoneyRequest : TGetMoneyRequestProc
        read FOnGetMoneyRequest
       write FOnGetMoneyRequest;
    property OnGetIncomingPaymentApproval : TGetIncomingPaymentApprovalProc
        read FOnGetIncomingPaymentApproval
       write FOnGetIncomingPaymentApproval;
    property OnPaymentAuthorizationResult : TPaymentAuthorizationResultProc
        read FOnPaymentAuthorizationResult
       write FOnPaymentAuthorizationResult;
    property OnActivate : TNotifyEvent read FOnActivate Write FOnActivate;
    property OnDeactivate : TNotifyEvent read FOnDeactivate Write FOnDeactivate;
    property OnError : TErrorEvent read FOnError write FOnError;
    property BankId : Cardinal read FBankID write FBankId;
    property AccountId : Int64 read FAccountId write FAccountId;
    property PaymentDelay : integer read FDelay write FDelay;
    end;

const
// 
  SWalletDLLName         = 'WalletConnect.DLL';
  SShopStartListen       = 'ShopStartListen';
  SShopStopListen        = 'ShopStopListen';
  SShopGetMem            = 'ShopGetMem';
  SShopFreeMem           = 'ShopFreeMem';
  SDateTimeToNanoTime    = 'DateTimeToNanoTime';
  SNanoTimeToDateTime    = 'NanoTimeToDateTime';
  SSystemTimeToNanoTime  = 'SystemTimeToNanoTime';
  SNanoTimeToSystemTime  = 'NanoTimeToSystemTime';

// 
  SWalletActive          = '    -  ';
  SWalletDllNotFound     = '   '+SWalletDllName;
  SWalletDllIsInvalid    = '  '+SWalletDLLName;
  SWalletInactive        = '    -   ';
  SOnlyOneInstance       = '      ';
  SCannotAllocateResources
                         = '   ';
  SMissingShopName       = '     ';
  SErrorStartListen      = '   ';
  STooManyPrices         = '  ';

function GetWord ( const Src : string ; Index : integer ;
  Delimeters : TSysCharSet): string;

function GetParam ( const Src, Param : string ): string;  

procedure StringToContract ( const Src : string ; Contract : TContract );

implementation

uses Math;

function GetWord ( const Src : string ; Index : integer ;
  Delimeters : TSysCharSet): string;
var
  C,I,S,L : integer; // Current,Index,Start,Length
begin
  S := 1; I := 0; C := 0; L := Length(Src);
  while I < L do
    begin
    Inc(I);
    If (Src[i] in Delimeters) then
      begin
      inc(C);
      If C=Index then S := I+1;
      If C>Index then
        begin
        Result := Copy(Src,S,I-S);
        exit;
        end;
      Continue;
      end;
    end;
  If C=index then
    begin
    Result := Copy(Src,S,I-S+1);
    end
  else
    Result := '';
end;

function GetParam ( const Src, Param : string ): string;
var
  i : integer;
  s,d : string;
begin
  i := 0;
  s := GetWord(Src,i,[';']);
  while s<>'' do
    begin
    d := GetWord(s,0,['=']);
    If AnsiCompareText(d,Param)=0 then
      Result := GetWord(s,1,['=']);
    inc(i);
    s := GetWord(Src,i,[';']);
    end;
end;

procedure StringToContract ( const Src : string ; Contract : TContract );
var
  S,D,N : string;
  c : integer;
begin
  c := 0; S := GetWord(Src,c,[';']);
  while S<>'' do
    begin
    N := Trim(UpperCase(getWord(S,0,['='])));
    D := Trim(GetWord(S,1,['=']));
    If N='ID' then
      begin
      Contract.Data := D;
      end;
    If N = 'PRICE' then
      try
        Contract.AddPrice(StrToInt(GetWord(D,0,[','])),
          StrToInt64(GetWord(D,1,[','])));
      except end;
    If N = 'DESC' then
      Contract.Description := D;
    If N = 'CONTRACT' then
      Contract.Text := D;
    If N = 'BANK' then
      try
        Contract.Bank := StrToInt(D);
      except end;
    If N = 'ACCOUNT' then
      try
        Contract.Account := StrToInt64(D);
      except end;
    Inc(c); S := GetWord(Src,c,[';']);
    end;
end;

//     .
//     :
//  - ReplyResource   
//   http:// -  url 
//   file:// -       html 
//      html 

procedure TPayCashShop.MakeReplyResource ( Source : string ; var Resource : TReplyResource);
var
  S,T : string;
  FS : TFileStream;
begin
  If FActive then
    begin
    Resource.ResourceCachePeriod := 0;
    Resource.ResourceFlags := 0;
    If Source='' then
      begin
      Resource.ResourceType := rtNoResource;
      Resource.ResourceSize := 0;
      Resource.ResourceBody := nil;
      exit;
      end;
    s := UpperCase(Copy(Trim(Source),1,7));
    if s='HTTP://' then
      begin
      Resource.ResourceType := rtURL;
      Resource.ResourceSize := Length(Source)+1;
      Resource.ResourceBody := ShopString(Source);
      exit;
      end;
    if S='FILE://' then
      begin
      Resource.ResourceType := rtHTML;
      //
      Delete(Source,1,7);
      try
        //     
        FS := TFileStream.Create(Source,fmOpenRead);
        SetLength(T,FS.Size);
        FS.Read(T[1],FS.Size);
        FS.Free;
        //    
        Resource.ResourceSize := Length(T)+1;
        Resource.ResourceBody := ShopString(T);
      except
        Resource.ResourceType := rtNoResource;
        Resource.ResourceSize := 0;
        Resource.ResourceBody := nil;
      end;
      exit;
      end;
    Resource.ResourceType := rtHTML;
    Resource.ResourceSize := Length(Source)+1;
    Resource.ResourceBody := ShopString(Source);
    end
  else
    raise TShopError.Create(SWalletInactive);
end;

var
  ShopInstance : TPayCashShop;

//     ,  

procedure LocalGetMoneyRequest(ErrorCode: TProcResult;
        Cypher: TCypher;
  const ContractModifier: TParamsRecord;
        PayerAddress: PChar;
  out RequestMode: word;
  out ContractData: TContractData;
  out ReplyResource: TReplyResource); stdcall;
var
  Param : TMoneyRequest;
begin
  if Assigned( ShopInstance ) then
    begin
    Param.ErrorCode        := ErrorCode;
    Param.Cypher           := Cypher;
    Param.ContractModifier := ContractModifier;
    Param.PayerAddress     := PayerAddress;
    //    
    EnterCriticalSection ( ShopInstance.FSection );
    ShopInstance.GetMoneyRequest(Param);
    LeaveCriticalSection( ShopInstance.FSection );
    RequestMode            := Param.RequestMode;
    ContractData           := Param.ContractData;
    ReplyResource          := Param.ReplyResource;
    end;
end;

procedure LocalGetIncomingPaymentApproval(ErrorCode: TProcResult;
  const PaymentInfo: TIncomingPaymentInfo;
  var ShortDescription:TShortDescription; // in/out
  out ConfirmAction: TConfirmAction;
  out BankConnection: DWord;
  out ReplyResource: TReplyResource); stdcall;
var
  Param : TIncomingPaymentApproval;
begin
  if Assigned( ShopInstance ) then
    begin
    Param.ErrorCode        := ErrorCode;
    Param.PaymentInfo      := PaymentInfo;
    Param.ShortDescription := ShortDescription;
    EnterCriticalSection(ShopInstance.FSection);
    ShopInstance.GetIncomingPaymentApproval(Param);
    LeaveCriticalSection(ShopInstance.FSection);
    ConfirmAction          := Param.ConfirmAction;
    BankConnection         := Param.BankConnection;
    ReplyResource          := Param.ReplyResource;
    end;
end;

procedure LocalPaymentAuthorizationResult( const PaymentAuthorizationInfo:
  TPaymentAuthorizationInfo; out ReplyResource: TReplyResource); stdcall;
var
  Param : TPaymentAuthorizationResult;
begin
  if Assigned( ShopInstance ) then
    begin
    Param.PaymentAuthorizationInfo := PaymentAuthorizationInfo;
    EnterCriticalSection(ShopInstance.FSection);
    ShopInstance.PaymentAuthorizationResult(Param);
    LeaveCriticalSection(ShopInstance.FSection);
    ReplyResource := Param.ReplyResource;
    end;
end;

{ TPayCashShop }

procedure TPayCashShop.Connect;
var
  DLLValid : boolean;
  Procs : TShopProcsTable;
begin
  if not Active then
    begin
    If ShopName='' then
      DoError(SMissingShopName);
    FWallet.Handle := LoadLibrary ( SWalletDLLName );
    If FWallet.Handle <> 0 then
      with FWallet do
        begin
        //   
        ShopStartListen := GetProcAddress( Handle, SShopStartListen );
        ShopStopListen  := GetProcAddress( Handle, SShopStopListen );
        ShopGetMem      := GetProcAddress( Handle, SShopGetMem );
        ShopFreeMem     := GetProcAddress( Handle, SShopFreeMem );
        DateTimeToNanoTime :=
          GetProcAddress( Handle, SDateTimeToNanoTime );
        NanoTimeToDateTime :=
          GetProcAddress( Handle, SNanoTimeToDateTime );
        SystemTimeToNanoTime :=
          GetProcAddress( Handle, SSystemTimeToNanoTime );
        NanoTimeToSystemTime :=
          GetProcAddress( Handle, SNanoTimeToSystemTime );
        DLLValid :=
          Assigned(ShopStartListen) and
          Assigned(ShopStopListen) and
          Assigned(ShopGetmem) and
          Assigned(ShopFreeMem) and
          Assigned(DateTimeToNanoTime) and
          Assigned(NanoTimeToDateTime) and
          Assigned(SystemTimeToNanoTime) and
          Assigned(NanoTimeToSystemTime);
        if not DLLValid then
          begin
          FreeLibrary(FWallet.Handle);
          DoError(SWalletDllIsInvalid);
          end;
        //   
        with Procs do
          begin
          GetMoneyRequest            := LocalGetMoneyRequest;
          GetIncomingPaymentApproval := LocalGetIncomingPaymentApproval;
          PaymentAuthorizationResult := LocalPaymentAuthorizationResult;
          end;
        if ShopStartListen(PChar(ShopName),Procs)=0 then
          begin
          FActive := true;             
          InitializeCriticalSection( FSection );
          DoActivate;
          end
        else
          begin
          FreeLibrary(FWallet.Handle);
          DoError(SErrorStartListen);
          end;
        end
    else
      DoError(SWalletDllNotFound);
    end;
end;

constructor TPayCashShop.Create(AOwner: TComponent);
begin
  If not assigned(ShopInstance) then
    begin
    inherited;
    FWallet.Handle := 0;
    FActive        := false;
    FDelay         := 60;     
    ShopInstance   := self;
    end
  else
    DoError(SOnlyOneInstance);
end;

destructor TPayCashShop.Destroy;
begin
  Disconnect;
  ShopInstance := nil;
  inherited;
end;

procedure TPayCashShop.Disconnect;
begin
  if FActive then
    try
      FWallet.ShopStopListen;
      DeleteCriticalSection(FSection);
      FreeLibrary(FWallet.Handle);
      FActive := false;
      DoDeactivate;
    except end;
end;

procedure TPayCashShop.DoActivate;
begin
  If Assigned (FOnActivate) then
    FOnActivate(Self);
end;

procedure TPayCashShop.DoDeactivate;
begin
  If Assigned (FOnDeactivate) then
    FOnDeactivate(Self);
end;

procedure TPayCashShop.DoError(Mess: string);
begin
  if Assigned (FOnError) then
    begin
    FOnError(Self,Mess);
    Abort;
    end
  else
    raise TShopError.Create(Mess);
end;

procedure TPayCashShop.FreeMem(Mem: Pointer; Size: integer);
begin
  If FActive then
    FWallet.ShopFreeMem(Mem,Size);
end;

function TPayCashShop.GetMem(Size: integer): Pointer;
begin
  If FActive then
    result := FWallet.ShopGetMem( Size )
  else
    result := nil;
end;

procedure TPayCashShop.SetActive(const Value: boolean);
begin
  If Value<>FActive then
    case Value of
      true : Connect;
      false : Disconnect;
      end;
end;

procedure TPayCashShop.SetShopName(const Value: string);
begin
  If not Active then FName := Value
  else
    DoError(SWalletActive);
end;

function TPayCashShop.ShopString(src: string): Pointer;
var
  C : integer;
begin
  if FActive then
    begin
    C := Length(Src)+1;
    result := GetMem(C);
    Move(Src[1],Result^,C);
    end
  else
    result := nil;
end;

procedure TPayCashShop.GetIncomingPaymentApproval
  (var M: TIncomingPaymentApproval);
var
  Param : TIncomingPaymentApproval;
var
  ID : string;
  RR : string;
begin
  if Assigned(FOnGetIncomingPaymentApproval) then
    with Param do
      begin
      SetLength(ID,PaymentInfo.PayeeRegDataSize);
      Move(PaymentInfo.PayeeRegData^,ID[1],
        PaymentInfo.PayeeRegDataSize);
      FOnGetIncomingPaymentApproval(ErrorCode,ID,ConfirmAction,RR);
      BankConnection := pmDefault;
      MakeReplyResource(RR,ReplyResource);
      end;
end;

procedure TPayCashShop.GetMoneyRequest(var M: TMoneyRequest);
var
  SL : TStringList;
  RR : string;
  i : integer;
  Contract : TContract;

  procedure ProcessContract;
  var
    s : string;
    i : integer;
  begin
    FillChar(M.ContractData,SizeOf(M.ContractData),0);
    with M.ContractData do
      begin
      If Contract.Data<>'' then
        begin
        PayeeRegDataSize := Length(Contract.Data)+1;
        PayeeRegData := ShopString(Contract.Data);
        end;
      S := Contract.Description+#0;
      Move(S[1],ShortDescription,
        Min(SizeOf(ShortDescription),Length(S)));
      PaymentDelay := FDelay;
      ProcessingOptions := proAcceptPaymentWithoutConfirmation;
      CurrencyOptionsCount := Contract.FPriceCount;
      For i := 0 to CurrencyOptionsCount-1 do
        begin
        CurrencyOptions[i].Currency := Contract.FPrices[i].Currency;
        CurrencyOptions[i].Sum := Contract.FPrices[i].Sum;
        end;
      MoneyDestination.BankID := Contract.Bank;
      MoneyDestination.AccountID := Contract.FAccount;
      ContractTermsFormat := ctf_Win1251;
      ContractTermsDataSize := Length(Contract.Text)+1;
      ContractTermsData := ShopString(Contract.Text);
      ContractSigner := M.Cypher;
      end;
  end;

begin
  If Assigned(FOnGetMoneyRequest) then
    with M do
      begin
      //  ,   
      SL := TStringList.Create;
      try
        For i := 0 to ContractModifier.ParamCount-1 do
          begin
          SL.Add(Format('%S=%S',[ContractModifier.ParamNames^[i],
            ContractModifier.ParamValues^[i]]));
          end;
        SL.Add(Format('PayerAddress=%s',[PayerAddress]));
        SL.Add(Format('Cypher=%X',[Cypher]));
        //  
        Contract := TContract.Create;
        Contract.Bank := BankId;
        Contract.Account := AccountId;
        FOnGetMoneyRequest(ErrorCode,SL,RequestMode,Contract,RR);
        //  
        ProcessContract;
        Contract.Free;
        MakeReplyResource(RR,ReplyResource);
      finally
        SL.Free;
      end;
    end;
end;

procedure TPayCashShop.PaymentAuthorizationResult
  (var M: TPaymentAuthorizationResult);
var
  ID : string;
  RR : string;
begin
  If Assigned(FOnPaymentAuthorizationResult) then
    with M do
      begin
      //    
      SetLength(ID,PaymentAuthorizationInfo.PayeeRegDataSize);
      Move(PaymentAuthorizationInfo.PayeeRegData^,ID[1],
        PaymentAuthorizationInfo.PayeeRegDataSize);
      //   
      FOnPaymentAuthorizationResult(PaymentAuthorizationInfo.ErrorCode,
        ID,RR);
      MakeReplyResource(RR,ReplyResource);
      end;
end;

{ TContract }

procedure TContract.AddPrice(Cur: TCurrencyCode; Price: TPreciseSum);
begin
  If FPriceCount>=MaxContractCurrencyOptionsCount then
    ShopInstance.DoError(STooManyPrices);
  FPrices[FPriceCount].Currency := Cur;
  FPrices[FPriceCount].Sum      := Price;
  Inc(FPriceCount);
end;

end.
